@@ -9,7 +9,6 @@ from tempfile import mkdtemp
 import mdtraj
 from simtk import openmm as mm
 from simtk.openmm import app
-from simtk.openmm import unit
 
 from .custom_reporters import FESReporter
 from .utils import *
@@ -57,7 +56,7 @@ class MDSimulation():
         >>> mdsim.run(1*unit.nanosecond)
         >>> mdsim.update_restraint_weight(2)
         >>> mdsim.run(1*unit.nanosecond)
-        >>> mdsim.remove_positional_restratins()
+        >>> mdsim.remove_positional_restraints()
         >>> end_parmed = mdsim.run(4 * unit.nanosecond)
 
         Parameters
@@ -263,7 +262,7 @@ class MDSimulation():
             strength of the positional restraint in kcal/mol/A^2
 
         reference_coordinates : unit.Quantity, optional
-            reference coordinates to be used with restraint defaults to current positions
+            reference coordinates to be used with restraint; defaults to current positions
 
         """
         print('setting restraints for {} atoms'.format(len(atoms_to_restrain)))
@@ -317,17 +316,13 @@ class MDSimulation():
         self.simulation.context.setPositions(self.positions)
 
     def set_implicit_solvent_model(self, implicit_solvent_model, nonbonded_method, nonbonded_cutoff):
-        """ Add an implicit solvent force to the system
+        """ Add an implicit solvent force to the system. OpenMM force index stored in self.custom_forces['implicit_solvent']
 
         Parameters
         ----------
 
         implicit_solvent_model : {'HCT', 'OBC1', 'OBC2', 'GBn', 'GBn2'}
 
-        nonbonded_method : {'PME'}
-
-        nonbonded_cutoff : unit.Quantity
-
         """
         if "implict_solvent" in self.custom_forces:
             raise ValueError('Implicit solvent model has already been set for this system')
@@ -336,20 +331,18 @@ class MDSimulation():
             raise ValueError('Unknown implicit solvent model {}'.format(implicit_solvent_model))
 
         implicit_model_object = eval("app.%s" % implicit_solvent_model)
-        if nonbonded_method is not None:
-            nonbonded_method_object = eval("app.%s" % nonbonded_method)
-            implicit_force = self.parmed_structure.omm_gbsa_force(implicit_model_object,
-                                                                  temperature=self.temperature,
-                                                                  nonbondedMethod=nonbonded_method_object,
-                                                                  nonbondedCutoff=nonbonded_cutoff)
-        else:
-            implicit_force = self.parmed_structure.omm_gbsa_force(implicit_model_object,
+
+        implicit_force = self.parmed_structure.omm_gbsa_force(implicit_model_object,
                                                                   temperature=self.temperature)
 
         self.custom_forces['implicit_solvent'] = self.system.addForce(implicit_force)
 
     def freeze_atom_selection(self, atoms_to_freeze):
-        """freeze selected atoms
+        """freeze selected atoms by setting their masses to zero. Stores masses and indeces in internal
+            dictionary (self.frozen_atoms[str(atom_index)]=mass.
+
+        atoms_to_freeze list of integers
+            atom indeces
 
         """
 
@@ -374,9 +367,14 @@ class MDSimulation():
         print('unfroze {} atoms'.format(len(self.frozen_atoms)))
         self.frozen_atoms = {}
 
-    def restrain_COM(self, atom_selection, COM_restraint_weight):
+    def restrain_COM(self, atom_selection, COM_restraint_weight=20):
         """ Adds a positional restraint to the COM of the atom_selection
 
+        atom_selection list of integers
+
+        COM_restraint_weight int
+            force constant to be used in COM restraint
+
         """
 
         # get the masses of the selection
@@ -408,6 +406,11 @@ class MDSimulation():
         self.system.removeForce(self.custom_forces['COM_restraint'])
 
     def center_coordinates(self):
+        """
+        translate the coordinates of the system so that the geometric center is at the origin
+
+        :return:
+        """
         print('centering coordinates')
         cog = np.mean(self.coordinates, axis=0)
         # System box vectors
@@ -419,8 +422,33 @@ class MDSimulation():
         new_coords = self.coordinates + delta
         self.positions = new_coords * unit.angstrom
 
+    def add_barostat(self):
+        if self.nonbonded_method is not None:
+
+            if self.membrane:
+                print('setting membrane barostat')
+                barostat_force = mm.MonteCarloMembraneBarostat(1 * unit.bar, 200 * unit.bar * unit.nanometer,
+                                                               self.temperature,
+                                                               mm.MonteCarloMembraneBarostat.XYIsotropic,
+                                                               mm.MonteCarloMembraneBarostat.ZFree)
+                self.custom_forces['barostat'] = self.system.addForce(barostat_force)
+
+            else:
+                print('setting barostat with pressure {}'.format(self.pressure))
+                barostat_force = mm.MonteCarloBarostat(self.pressure, self.temperature, 25)
+                self.custom_forces['barostat'] = self.system.addForce(barostat_force)
+        else:
+            print('cant use barostat with non-periodic system, overriding constant pressure condition')
+
+    def remove_barostat(self):
+        self.system.removeForce(self.custom_forces['barostat'])
+        self.simulation.context.reinitialize()
+        self.simulation.context.setPositions(self.positions)
+        self.simulation.context.setPeriodicBoxVectors(self.box_vectors[0], self.box_vectors[1], self.box_vectors[2])
+
     def _build_sim(self):
-        """Use params specified on init to build the simulation context, specify the platform and bind an integrator
+        """Use params specified on init to build the simulation context, set the cwd,
+        specify the platform and bind an integrator
 
         """
 
@@ -460,22 +488,7 @@ class MDSimulation():
             self.set_implicit_solvent_model(self.implicit_solvent_model, self.nonbonded_method, self.nonbonded_cutoff)
 
         if self.pressure:
-            if self.nonbonded_method is not None:
-
-                if self.membrane:
-                    print('setting membrane barostat')
-                    barostat_force = mm.MonteCarloMembraneBarostat(1 * unit.bar, 200 * unit.bar * unit.nanometer,
-                                                                   self.temperature,
-                                                                   mm.MonteCarloMembraneBarostat.XYIsotropic,
-                                                                   mm.MonteCarloMembraneBarostat.ZFree)
-                    self.system.addForce(barostat_force)
-
-                else:
-                    print('setting barostat with pressure {}'.format(self.pressure))
-                    barostat_force = mm.MonteCarloBarostat(self.pressure, self.temperature, 25)
-                    self.system.addForce(barostat_force)
-            else:
-                print('cant use barostat with non-periodic system, overriding constant pressure condition')
+            self.add_barostat()
 
         if self.atoms_to_freeze is not None:
             self.freeze_atom_selection(self.atoms_to_freeze)
@@ -533,15 +546,19 @@ class MDSimulation():
             self.restrain_COM(atoms, 20)
 
     def _build_custom_forces(self):
+        """Function for inherited classes"""
         pass
 
     def _write_coordinates_to_h5(self, fname='current_frame.h5'):
         """Writes minimized coordinates to an H5 file
 
+        fname string
+            name of minimized coords saved to hdf5 file. Should end in .h5 or .hdf5
+
         """
 
-        if not fname.endswith('h5'):
-            print('WARNING: saving an h5 file without the suffix .h5')
+        if not fname.endswith('h5') or fname.endswith('hdf5'):
+            print('WARNING: saving an h5 file without the suffix .h5 or .hdf5')
 
         # write hdf5 single frame
         current_state = self.simulation.context.getState(getPositions=True,enforcePeriodicBox=self.wrap_state_coords)
@@ -565,7 +582,10 @@ class MDSimulation():
 
     def pickle_sim(self, pickle_name='mdsim.pickle'):
         """ Remove the Swigpy objects that don't pickle and write the current simulation pickle to cwd
-        In order to use the pickle you will need to run _build_sim
+        In order to use the pickle you will need to run _build_sim and _build_custom_forces
+
+        pickle_name string
+            name of file to write pickle to. cwd is appended to the prefix.
 
         Examples
         -------
@@ -591,7 +611,7 @@ class MDSimulation():
         with open(fname, 'wb') as handle:
             pickle.dump(obj=self, file=handle)
         print('pickled sim saved as {}'.format(fname))
-        print('to use it once unpickled you will need to run self._build_sim()')
+        # print('to use it once unpickled you will need to run self._build_sim()')
 
     def purge_sim_files(self):
         """Removes all files in mdsim.cwd
@@ -614,9 +634,12 @@ class MDSimulation():
         del self.simulation
         self._build_sim()
 
-    def _add_reporters(self, total_steps):
+    def _add_reporters(self, total_steps=1):
         """ Add reporters to the sim object
 
+        total_steps int
+            total_steps is needed for accurate estimate of time remaining in progress reporter.
+
         """
         if self.progress_interval:
             if isinstance(self.progress_interval, unit.Quantity):
@@ -695,8 +718,8 @@ class MDSimulation():
             self.simulation.reporters.append(traj_reporter)
 
     def update_parmed(self):
-        """ updates the parmed object with velocities, coordinates, posiitons, and box vectors
-
+        """ updates the parmed object with velocities, coordinates, positions, box vectors,
+            coordinates and current potential energy are updated to the MDSimulation object
 
         """
         state = self.simulation.context.getState(getPositions=True,
@@ -707,6 +730,7 @@ class MDSimulation():
         current_positions = state.getPositions()
         current_velocities = state.getVelocities()
         current_box = state.getPeriodicBoxVectors()
+        self.energy = state.getPotentialEnergy()
 
         self.coordinates = np.array(current_positions.value_in_unit(unit.nanometer))
         self.parmed_structure.positions = current_positions
@@ -715,7 +739,7 @@ class MDSimulation():
             self.parmed_structure.box_vectors = current_box
         self.positions = current_positions
 
-    def minimize(self, max_steps=None, save_h5=False):
+    def minimize(self, max_steps=None, save_h5=False, h5_file='minimized.h5'):
         """Minimize the energy of the current simulation
 
         :param max_steps: int, optional
@@ -746,7 +770,7 @@ class MDSimulation():
         print('successfully minimized the system to final energy {}'.format(final_energy))
 
         if save_h5:
-            self._write_coordinates_to_h5(fname='minimized.h5')
+            self._write_coordinates_to_h5(fname=h5_file)
 
         current_box = state.getPeriodicBoxVectors()
 
@@ -795,23 +819,50 @@ class GentleHeatSimulation(MDSimulation):
 
     :start_temp: unit.Quantity
     :target_temp: unit.Quantity
-    :start_restraint_weight unit.Quantity or float
-    :num_heat_stages int
-    :num_release_stages
+    :start_restraint_weight: unit.Quantity or float
+    :num_heat_stages: int
+    :num_release_stages: int
 
 
     """
     def __init__(self, start_temp=100,
                  start_restraint_weight=2, num_heat_stages=2,
                  num_release_stages=2, equil_time=2,
-                 stage_run_time=.5, **kwargs):
+                 stage_time=.5,
+                 heat_pressure=None, release_pressure=None, equil_pressure=None,
+                 **kwargs):
         super().__init__(**kwargs)
         self.start_temp = start_temp
         self.start_restraint_weight = start_restraint_weight
         self.num_heat_stages = num_heat_stages
         self.num_release_stages = num_release_stages
+        self.heat_pressure = heat_pressure
+        self.release_pressure = release_pressure
+        self.equil_pressure = equil_pressure
         self.equil_time = equil_time
-        self.stage_run_time = stage_run_time
+        self.stage_time = stage_time
+
+    def toggle_pressure(self, desired_state, pressure_on):
+        if desired_state=='ON':
+            if pressure_on:
+                pass
+            else:
+                state = self.simulation.context.getState(getPositions=True,
+                                                         enforcePeriodicBox=self.wrap_state_coords)
+                self.positions = state.getPositions()
+                self.add_barostat()
+                self.simulation.context.reinitialize()
+                self.simulation.context.setPositions(self.positions)
+                self.simulation.context.setPeriodicBoxVectors(self.box_vectors[0], self.box_vectors[1], self.box_vectors[2])
+                pressure_on=True
+        elif desired_state=='OFF':
+            if not pressure_on:
+                pass
+            else:
+                print('Turning off barostat')
+                pressure_on = False
+                self.remove_barostat()
+        return pressure_on
 
     def run(self):
         """
@@ -820,23 +871,39 @@ class GentleHeatSimulation(MDSimulation):
         :return: parmed_structure
         """
 
+        if 'positional_restraints' not in self.custom_forces:
+            raise ValueError('Gentle Heat Simulations require at least one positional restraint')
+
+
+
+
         current_weight = self.start_restraint_weight
         current_temp = self.start_temp
         temp_interval = (self.temperature.value_in_unit(unit.kelvin)-self.start_temp)/self.num_heat_stages
         if self.equil_time is not None:
             total_steps = ((self.num_heat_stages + self.num_release_stages) \
-                           * self.stage_run_time + self.equil_time) * unit.nanosecond / self.step_length
+                           * self.stage_time + self.equil_time) * unit.nanosecond / self.step_length
             equil_steps = int(round(self.equil_time * unit.nanosecond /self.step_length))
         else:
             total_steps = ((self.num_heat_stages + self.num_release_stages) \
-                           * self.stage_run_time) * unit.nanosecond / self.step_length
+                           * self.stage_time) * unit.nanosecond / self.step_length
 
         self._add_reporters(total_steps)
 
         print('starting temp', self.start_temp)
-        print('starting restraint weight ',current_weight)
 
-        stage_steps = int(round(self.stage_run_time * unit.nanosecond /self.step_length))
+        stage_steps = int(round(self.stage_time * unit.nanosecond /self.step_length))
+
+        #add or remove the temperature depending on input bools
+        if 'barostat' in self.custom_forces:
+            pressure_on = True
+        else:
+            pressure_on = False
+
+        if self.heat_pressure:
+            pressure_on = self.toggle_pressure(desired_state='ON',pressure_on=pressure_on)
+        else:
+            pressure_on =self.toggle_pressure(desired_state='OFF', pressure_on=pressure_on)
 
         for i in range(self.num_heat_stages):
             self.simulation.step(stage_steps)
@@ -844,12 +911,25 @@ class GentleHeatSimulation(MDSimulation):
             print(' setting temp to ',current_temp)
             self.integrator.setTemperature(current_temp * unit.kelvin)
 
+        if self.release_pressure:
+            pressure_on = self.toggle_pressure(desired_state='ON', pressure_on=pressure_on)
+        else:
+            pressure_on = self.toggle_pressure(desired_state='OFF', pressure_on=pressure_on)
+
+
+        print('beginning release stage with starting restraint weight ',current_weight)
+
         for i in range(self.num_release_stages):
             self.simulation.step(stage_steps)
-            print('setting restraint weight to ',current_weight)
             current_weight -= self.start_restraint_weight/self.num_release_stages
+            # print('setting restraint weight to ',current_weight)
             self.update_restraint_weight(current_weight)
 
+        if self.equil_pressure:
+            pressure_on = self.toggle_pressure(desired_state='ON', pressure_on=pressure_on)
+        else:
+            pressure_on = self.toggle_pressure(desired_state='OFF', pressure_on=pressure_on)
+
         if self.equil_time is not None:
             print('equilibrating for {} ns'.format(self.equil_time))
             self.simulation.step(equil_steps)
@@ -865,7 +945,18 @@ class MetadynamicSimulation(MDSimulation):
     Metadynamics Simulation Class
 
     A maximum of 3 CVs can be specified, and each CV must be one of the following:
-    'Distance, RMSD, Z-Proj, Angle'
+    'Distance, RMSD, Z-Proj, Angle'. Each CV must have exactly one value specified in the following format:
+    [CV_name (string), mask_1 (string), mask_2 (string), CV_min (float), CV_max (float), CV_bin_length (float)]
+
+    Masks follow rs_simtools select atoms syntax where masks beginning with 'resid' are assumed to be resid_selection
+    keywords and masks without 'resid' are assumed to be keyword_selection inputs. Where only one
+    mask is needed (i.e. for RMSD or Angle) the second mask is ignored, and should be specified as 'none'. CV dimensions
+    should be constructed such that all CVs have the same discrete number of bins, even if the units on those bins
+    are not equal.
+
+    All structures should be aligned such that the Distance CV is aligned along the Z-axis. This can be done using
+    rs_simtools align_structure() command, but should be done before solvating the system as rotating a system after
+    solvation can lead to bad periodic boundary conditions.
 
     Examples
     --------
@@ -876,8 +967,9 @@ class MetadynamicSimulation(MDSimulation):
     >>> ['RMSD', 'protein', 'None', 0, 2, .02 ]
     >>> ]
 
+
     >>> metaD = MetadynamicSimulation(parmed_structure = parmed_structure,
-    >>>                        dome=True,
+    >>>                        funnel=True,
     >>>                        CV_list= CV_list, run_time=25 * unit.nanosecond)
     >>>
     >>>metaD.run()
@@ -891,6 +983,12 @@ class MetadynamicSimulation(MDSimulation):
 
     dome : bool, optional
 
+    anchor_selection: string
+        ommtk selection syntax for where to anchor the flat bottom against. Will use center of geometry.
+
+    fb_scale_factor : float
+        fb restraint radius will automatically be set to the max distance on protein from ligand * this factor
+
     radius_scale_factor : float, optional
         factor for determining the radius of funnel or dome relative to ligand size
 
@@ -918,13 +1016,14 @@ class MetadynamicSimulation(MDSimulation):
 
     def __init__(self, CV_list=[],
                  funnel=False, radius_scale_factor=2,
-                 dome=False, funnel_angle=.5,
+                 dome=False, anchor_selection=None, funnel_angle=.5,
                  bias_factor=20,
                  initial_bias_height=2 * unit.kilojoule_per_mole,
                  bias_frequency=1 * unit.picoseconds,
                  bias_save_frequency=1 * unit.picoseconds,
                  bias_directory=None,
                  fes_interval=.5 * unit.nanoseconds,
+                 fb_scale_factor=.8,
                  **kwargs):
 
         super().__init__(**kwargs)
@@ -941,6 +1040,8 @@ class MetadynamicSimulation(MDSimulation):
         self.funnel = funnel
         self.radius_scale_factor = radius_scale_factor
         self.funnel_angle = funnel_angle
+        self.anchor_selection = anchor_selection
+        self.fb_scale_factor = fb_scale_factor
 
         self._build_custom_forces()
 
@@ -953,46 +1054,50 @@ class MetadynamicSimulation(MDSimulation):
 
 
         # build funnel or dome restraints
-        #TODO try using a customcentroidbond force with flat bottom for this
         if self.dome:
+
+            if self.anchor_selection == None:
+                raise ValueError('Cant use flat bottom restraint without anchor_selection')
+
             if self.membrane:
                 raise ValueError(
-                    'Spherical restraints should not be used for channel protein systems with deep binding pockets.')
+                    'Flat bottom restraint should not be used for channel protein systems with deep binding pockets.')
 
             # Now get the setup stage for placing sphere (otherwise coords may be wrapped)
             lig_atom_index = select_atoms(self.parmed_structure, keyword_selection='ligand')
             prot_atom_index = select_atoms(self.parmed_structure, keyword_selection='protein')
 
             lig_positions = np.array(self.positions.value_in_unit(unit.nanometer))[lig_atom_index]
+            prot_positions = np.array(self.positions.value_in_unit(unit.nanometer))[prot_atom_index]
 
             if len(lig_atom_index) == 0 or len(prot_atom_index) == 0:
-                raise ValueError('could not select ligand or protein for dome placement')
+                raise ValueError('could not select ligand or protein for flat bottom restraint')
 
-            radius, center = find_radius_and_center(lig_positions, self.radius_scale_factor)
+            #use this function to estimate the radius of the flat bottom pot
+            radius = find_flat_bottom_radius(prot_positions,lig_positions, self.fb_scale_factor)
 
-            print('Adding a spherical potential to keep ligand near binding site ',
-                  'using the following params',
-                  ' radius: {} rcenter: {}'.format(radius, center))
-            print('Note, this should only be used for shallow binding pockets')
+            print('Adding a flat-bottom restraint with radius: {} '.format(radius))
 
-            spherical_boundary_force = mm.CustomExternalForce(
-                'K*max(0, (r-radius))^2; r=sqrt(periodicdistance(x,y,z,x0,y0,z0)^2)')
+            fb_force = mm.CustomCentroidBondForce(2, 'step(r-r0) * (fb_k) * (r-r0)^2; r=distance(g1,g2)')
+            fb_force.addGlobalParameter('r0',radius)
+            fb_force.addGlobalParameter('fb_k',20)
+            # find anchor group
+            if self.anchor_selection.startswith('resid'):
+                resid_list = self.anchor_selection[6:]
+                anchor_group = select_atoms(self.parmed_structure, resid_selection=resid_list)
+            else:
+                anchor_group = select_atoms(self.parmed_structure, keyword_selection=self.anchor_selection)
 
-            spherical_boundary_force.addGlobalParameter('K', 30)
-            spherical_boundary_force.addGlobalParameter('radius', radius)
-            spherical_boundary_force.addGlobalParameter('x0', center[0])
-            spherical_boundary_force.addGlobalParameter('y0', center[1])
-            spherical_boundary_force.addGlobalParameter('z0', center[2])
 
-            self.custom_forces['dome'] = self.system.addForce(spherical_boundary_force)
+            if len(anchor_group) == 0:
+                raise ValueError('Anchor group atom selection set has no atoms')
 
-            print('applying dome to {} num of atoms'.format(len(lig_atom_index)))
-            for i in lig_atom_index:
-                spherical_boundary_force.addParticle(i, [])
-
-            # use the inherited restrain COM to keep the protein from floating away
-            self.restrain_COM(atom_selection=prot_atom_index, COM_restraint_weight=30)
+            print('num atoms being used in anchor moeity: {}'.format(len(anchor_group)))
+            fb_force.addGroup(anchor_group)
+            fb_force.addGroup(lig_atom_index)
+            fb_force.addBond([0, 1])
 
+            self.custom_forces['flat_bottom_restraint'] = self.system.addForce(fb_force)
 
         if self.funnel:
             print ('Warning! Funnel should only be used for z-axis aligned structures.')
@@ -1250,20 +1355,21 @@ class MetadynamicSimulation(MDSimulation):
 class SimulatedTemperingSimulation(MDSimulation):
     """
     Simulated Tempering simulation which samples multiple temperatures via a Weng-Landau weights and
-    metropolis monte carlo temperature updates.
+    metropolis monte carlo temperature updates. The temperatures are specified with minTemperature and
+    maxTemperature and numTemperatures where temperatures are evenly spaced across the range.  
 
     Parameters
     ----------
 
-    minTemperature:
+    minTemperature: float
 
-    numTemperatures:
+    numTemperatures: int
 
-    maxTemperature:
+    maxTemperature: float
 
-    reportInterval:
+    reportInterval: int
 
-    reportFile:
+    reportFile: string
 
     """
     def __init__(self, numTemperatures=7, minTemperature=300 * unit.kelvin, maxTemperature=450 * unit.kelvin,
